/*
  This file defines an "optimised" data-structure to store factors of 
  sigma(p^q). Note that nothing has to be verified here. It just acts as an
  "oracle" which gives possible factors, and the factors are verified in 
  "oraclefact()" in the main file.
*/

class facts_t {
private:
  set<int_t> *facts;

  const facts_t &operator=(const facts_t &f) {
    return *this;
  }

public:
  static int nbfacts;
  static int nbfacts_set;

  static void facts_stats() {
    printf("facts_t : %d / %d\n",nbfacts,nbfacts_set);
  }

  unsigned long long ssf;
  unsigned int sf[4];
  
  facts_t() {ssf=0;facts=NULL;nbfacts++;bzero(sf,sizeof(sf));}
  facts_t(const facts_t &f):facts(f.facts),ssf(f.ssf) {nbfacts++;memcpy(sf,f.sf,sizeof(sf));if(facts) {facts=new set<int_t>(*(f.facts));nbfacts_set++;}}
    
  ~facts_t() {if(facts) {delete facts;nbfacts_set--;} nbfacts--;}
  
  void setused() {}

  bool insert(const int_t &f) {
    if(f<int_t(B3)) {
      unsigned int fi=toint(f);
      if(fi<B3) {
	int ii=rprimes[fi];
	assert(ii>=0);
	if(ii<64) {
	  if(ssf&(1LL<<ii)) return 0; 
	  ssf|=(1LL<<ii); 
	  return 1;
	}
      }
      for(int i=0;i<4;i++) {
	if(sf[i]==fi) return 0;
	if(sf[i]==0) {
	  sf[i]=fi;
	  return 1;
	}
      }
    }
    
    if(facts==NULL) {facts=new set<int_t>();nbfacts_set++;}
    
    if(facts->find(f)==facts->end()) {
      facts->insert(f);
      return 1;
    }
    return 0;
  }

  void add(const int_t &a) {insert(a);}
  void add(const set<int_t> &s) {
    set<int_t>::const_iterator it;
    for(it=s.begin();it!=s.end();it++) 
      add(*it);
  }

  struct const_iterator {
    const facts_t *f;
    set<int_t>::const_iterator it;
    int pos;
    
    const_iterator() {f=NULL;pos=0;}
    const_iterator(const facts_t &ff,int p=0) {f=&ff;pos=p;}
    
    void inc() {
      if(pos<1000) {
	pos++;
	if(pos<=64) {
	  if(pos==64 || (f->ssf>>pos)==0) pos=100;
	}
	if(pos==104) 
	  if(f->facts) {
	    pos=1000;it=f->facts->begin();
	    if(it==f->facts->end()) pos=10000;
	  }
	  else pos=10000;
	return;
      }
      it++;
      assert(f->facts);
      if(it==f->facts->end()) pos=10000;
    }

    bool isgood() {
      if(pos<64) return (f->ssf&(1LL<<pos))!=0;
      if(pos>=100 && pos<104) return f->sf[pos-100];
      if(pos==1000 && it!=f->facts->end()) return true;
      if(pos==10000) return true;
      assert(0);
    }
    
    void operator ++() {
      inc();
      while(!isgood()) inc();
    }

    bool operator!=(const const_iterator &i) const { return pos!=i.pos || (pos==1000 && it!=i.it);}
    
    const int_t operator *() const {
      if(pos<64) return primes[pos];
      if(pos<104 && pos>=100) return f->sf[pos-100];
      if(pos==1000) return *it;
      assert(0);
    }
  };
  
  const_iterator begin() const{ const_iterator c(*this,0);while(!c.isgood()) c.inc();return c;}
  const_iterator end() const { return const_iterator(*this,10000);}
  
};

int facts_t::nbfacts=0;
int facts_t::nbfacts_set=0;


struct sigmafacts_t {
  facts_t *factsmall[10];
  map<int,facts_t> *listfactors;
 
  sigmafacts_t() {
    listfactors=NULL;
    //factsmall=NULL;
    bzero(factsmall,sizeof(factsmall));
  }

  ~sigmafacts_t() {
    if(listfactors) delete listfactors;
    /*
    if(factsmall) {
      delete [] factsmall;
    }
    */
  }
};

//#define LISTHASH // some stranges things happens with hashtable (due to colisions...)

#ifdef LISTHASH
#define HASHSIZE 50000001

unsigned int hash(const int_t &p)
{
  return mpz_fdiv_ui(p.get_mpz_t(),HASHSIZE);
}
#endif

facts_t listfactorsss[30][30];

sigmafacts_t **sigmafactssmall;

#ifdef LISTHASH
sigmafacts_t *sigmafactsbig[HASHSIZE];
#else
map<pair<int_t,int>,facts_t > listfactorsbig;
//map<int_t,sigmafacts_t> listfactorsbig;
#endif

int lf_rprimes[100];

facts_t *getfacts(sigmafacts_t *sf, int q) {
  int iq=-1;
  if(q<100)
    iq=lf_rprimes[q];
  if(iq>=0 && iq<10) {
    facts_t *f=sf->factsmall[iq];
    if(f==NULL) f=sf->factsmall[iq]=new facts_t;
    return f;
  }
  if(sf->listfactors==NULL) sf->listfactors=new map<int,facts_t>;
  return &((*(sf->listfactors))[q]);
}

facts_t *getfacts(const int_t &p, int q) {
  if(p<30 && q<30) {
    unsigned int ip=toint(p);
    return &(listfactorsss[ip][q]);
  }
  if(p<B1) {
    unsigned int ip=toint(p);
    sigmafacts_t *sf=sigmafactssmall[ip];
    if(sf==NULL) sigmafactssmall[ip]=sf=new sigmafacts_t();
    return getfacts(sf,q);
  }

#ifdef LISTHASH
  unsigned int h=hash(p);
  sigmafacts_t *sf=sigmafactsbig[h];
  if(sf==NULL) sigmafactsbig[h]=sf=new sigmafacts_t();
  return getfacts(sf,q);
#else 
  map<pair<int_t,int>,facts_t >::iterator it;
  it=listfactorsbig.find(make_pair(p,q));
  if(it==listfactorsbig.end()) {
    return &listfactorsbig[(make_pair(p,q))];
  }
  return &(it->second);
#endif
}

void initlistfactors()
{
  sigmafactssmall=new sigmafacts_t*[B1];

  bzero(sigmafactssmall,sizeof(sigmafactssmall[0])*B1);

#ifdef LISTHASH
  bzero(sigmafactsbig,sizeof(sigmafactsbig[0])*HASHSIZE);
#endif

  int k=0;
  for(int i=1;i<100;i++) {
    if(isprimei(i+1)) {
      lf_rprimes[i]=k++;
    } else 
      lf_rprimes[i]=-1;
  }
}


bool addfact_nocheck(const int_t &p,int q,const int_t &f)
{
  facts_t *fa=getfacts(p,q-1);
  return fa->insert(f);
}

void makesmallfacts(const int_t &p,int q) 
{
  int_t s=sigma(p,q);
  if(isprime(s)) return;

  int i=0;
  while(primes[i]<2) i++;

  while(s>1) {
    if(s<B1) {
      int s_=toint(s);
      while(s_>1 && firstprimes[s_]!=0) {
	addfact_nocheck(p,q+1,firstprimes[s_]);
	s_/=firstprimes[s_];
      }
      if(s_>1) {
	addfact_nocheck(p,q+1,s_);
      }
      return;
    }
    for(;i<nbrprimes && int_t(primes[i])*primes[i]<s;i++) {
      if(isdiv(s,primes[i])) {
	addfact_nocheck(p,q+1,primes[i]);
	while(isdiv(s,primes[i]))
	  s/=primes[i];
	break;
      } 
    }
    assert(i!=nbrprimes);
  }
}

void addfact_(const int_t &p,int q,const int_t &f)
{
  int_t s=sigma(p,q-1);
  if(f>1 && isprime(p) && p!=2 && isprime(f) && isdiv(s,f)
     && isprimei(q)
     ) {
    //if(isprimei(q))
    addfact_nocheck(p,q,f);
  } else {
    cout<<f<<" bad fact of "<<p<<" "<<q<<" ";
    if(!isdiv(s,f)) cout<<"not-div ";
    if(!isprime(f)) cout<<"f-not-prime ";
    if(!isprime(p)) cout<<"p-not-prime ";
    if(!isprimei(q)) cout<<"q-not-prime ";    
    cout<<endl;
  }
}

void readfactorsp(const char *fn) 
{
  printf("readp %s\n",fn);
  FILE *in;
  in=fopen(fn,"r");
  if(!in) {printf("file %s not found\n",fn);return;}
  int q=0;
  char bp[100000];
  char bf[100000];
  int_t p=0,s=1;
  while(fgets(bf,100000,in)) {
    int i;
    for(i=1;bf[i]>='0'&&bf[i]<='9';i++);
    bf[i]=0;
    if(bf[0]=='p') 
      p=int_t(bf+1);
    if(bf[0]=='q') {
      assert(1==sscanf(bf+1,"%d",&q));
    }
    if(bf[0]>='0'&&bf[0]<='9') {
      int_t f(bf);
      addfact_(p,q,f);
    }
  }
  fclose(in);
  printf("%s readed\n",fn);
}

void readfactors_old(const char *fn) 
{
  printf("read %s\n",fn);
  FILE *in;
  in=fopen(fn,"r");
  if(!in) {printf("file %s not found\n",fn);return;}
  int q;
  char bp[100000];
  char bf[100000];
  char bb[100000];
  while(fgets(bb,100000,in)) {
    if(bb[0]=='#') continue;
    if(3!=sscanf(bb,"%s %d %s",bp,&q,bf)) continue;
    int_t p(bp);
    int_t f(bf);
    addfact_(p,q,f);
  }
  fclose(in);
  printf("%s readed\n",fn);
}

void readfactors(const char *fn) 
{
  FILE *in;
  in=fopen(fn,"r");
  if(!in) {printf("file %s not found\n",fn);return;}
  char bf[100000];
  int type=0;
  while(1) {
    if(fgets(bf,100000,in)==0) break;
    if(bf[0]==0 || bf[0]=='#') continue;
    if(bf[0]=='p') {type=2;break;}
    if(bf[0]>='0' && bf[0]<='9') {type=1;break;}
    break;
  }
  fclose(in);
  if(type==1) readfactors_old(fn);
  else if(type==2) readfactorsp(fn);
  else {
    printf("unknown file type %s\n",fn);
  }
}


void readfactors2(const char*fn="alouf.txt") 
{
  printf("read2 %s\n",fn);
  FILE *in;
  in=fopen(fn,"r");
  if(!in) {printf("file not found\n");return;}
  int_t p=0,f;int q;
  char bf[100000];
  char bf2[100000];
  while(1) {
    if(fgets(bf,100000,in)==0) break;
    strcpy(bf2,bf);
    //printf("%s\n",bf);
    if(bf[0]=='=') {
      int i;
      for(i=2;bf[i]>='0' && bf[i]<='9';i++);
      assert(i!=2);
      bf[i]=0;
      p=bf+2;
      //cout << "p= " << p << endl;
    } else {
      if(0==sscanf(bf,"%d",&q)) break;
      //cout << " q= " << q << endl;
      int i=0,j=0;
      int_t s=sigma(p,q-1);
      while(bf[i]!='|' && bf[i]) i++;
      assert(bf[i]);
      i++;
      while(bf[i] && bf[i]==' ') i++;
      if(bf[i]==0 || bf[i]=='\n') continue;
      while(1) {
	if(bf[i]>='0' && bf[i]<='9') {
	  int ii=i;
	  while(bf[i]&&(bf[i]>='0' && bf[i]<='9')) i++;
	  bool end=(bf[i]==0);
	  bf[i]=0;j=i+1;
	  while(bf[ii]=='0') ii++;
	  f=bf+ii;
	  //cout << "  f= "<< f << endl;
	  if(!isprime(f)) {
	    if(f>1000) {
	      cout << "warn fact not prime " << p << " " << q << " " << f <<endl;
	      if(s%f==0) 
		cout << "but OK\n";
	      printf(":%s\n",bf2);
	    }
	  } else if(s%f==0) {
	    addfact_(p,q,f);
	  } else {
	    if(f>1000) {
	      cout << "warn bad fact " << p << " " << q << " " << f <<endl;
	      printf(":%s\n",bf2);
	    }
	  }
	  if(end) break;
	  i++;
	  while(bf[i] && (bf[i]==' ' || bf[i]=='x' || bf[i]=='^')) i++;
	  if(bf[i]==0 || bf[i]=='\n') break;
	} else  if(bf[i]=='C') {
	  i++;
	  while(bf[i]&&(bf[i]>='0' && bf[i]<='9')) i++;
	  if(bf[i]==0 || bf[i]=='\n') break;
	  while(bf[i] && (bf[i]==' ' || bf[i]=='x')) i++;
	  if(bf[i]==0 || bf[i]=='\n') break;
	} else {
	  cout << "bad parse : '"<< bf <<"'\n"; 
	  break;
	}
      }
    }
  }
  fclose(in);
  printf("%s readed\n",fn);
}

